;;; From p. 67 Paul Graham "ANSI Common LISP"
(defun tokens (str &optional (test #'constituent) (start 0))
  (let ((p1 (position-if test str :start start)))
    (if p1
        (let ((p2 (position-if #'(lambda (c)
                                   (not (funcall test c)))
                               str :start p1)))
          (cons (subseq str p1 p2)
                (if p2
                    (tokens str test p2)
                      nil)))
      nil)))

;;; From p. 67 Paul Graham "ANSI Common LISP"           
(defun constituent (c)
  (and (graphic-char-p c)
       (not (char= c #\ ))))

;;; p. 164 Paul Graham "ANSI Common LISP"           
(defmacro while (test &rest body)
  `(do ()
       ((not ,test))
     ,@body))

;;; p 191 Paul Graham "On Lisp"
(defmacro aif (test-form then-form &optional else-form)
  `(let ((it ,test-form))
     (if it ,then-form ,else-form)))

(defun force-list (e)
  (if (listp e) e (list e)))

(defun copy-board (b)
  (let ((r (make-array (list x-max y-max))))
    (dotimes (x x-max)
      (dotimes (y y-max)
	(setf (aref r x y) (aref b x y))))
    r))
    


(defparameter *board* nil)
(defparameter *score* 0)
(defparameter x-max 0)
(defparameter y-max 0)

(defparameter *my-id* nil)
(defparameter *my-carrying-capacity* nil)
(defparameter *my-money* nil)
(defparameter *my-pos* nil)
(defparameter *my-route* nil)

(defconstant *tile-plain* #\.)
(defconstant *tile-lethal* #\~)
(defconstant *tile-wall* #\#)
(defconstant *tile-home* #\@)


(defparameter *base-list* nil)

(defun get-bases (&optional (board *board*))
  (dotimes (x x-max)
    (dotimes (y y-max)
      (when (equal (get-board (list x y) board) *tile-home*) (push (list x y) *base-list*)))))

(defun getx (pos) (elt pos 0))
(defun gety (pos) (elt pos 1))

(defun move (pos dir)
  (case dir
    ('N (list (getx pos) (+ (gety pos) 1)))
    ('S (list (getx pos) (- (gety pos) 1)))
    ('E (list (+ (getx pos) 1) (gety pos)))
    ('W (list (- (getx pos) 1) (gety pos)))))
    
;;; Notes:
;;;
;;; If a robot dies, his dropped packages can not be retrieved by someone else.
;;;
;;; TODO -- need to infer when robots die (by the fact they aren't listed)

(defstruct pkg
  reported-pos
  held-by
  dest-pos
  weight)

(defun apply-moves (start moves)
  (let ((result nil)
	(pos start))
    (dolist (dir moves)
      (push pos result)
      (setf pos (move pos dir)))
    (reverse result)))

(defun make-route (pos dest &optional (solver #'better-solver))
  (let* ((moves (funcall solver pos dest))
	 (coordinates (apply-moves pos moves)))
    (list dest moves coordinates)))


(defun route-dest (r)
  (first r))

(defun route-moves (r)
  (second r))

(defun route-coordinates (r)
  (third r))

(defun continue-route ()
  (let* ((c (route-coordinates *my-route*))
	 (p (position *my-pos* c :test #'equal :from-end t)))
    (if (and p (not (equal *my-pos* (route-dest *my-route*))))
	(move-myself (elt (route-moves *my-route*) p))
      (progn (setf *my-route* nil)
	     "Pick 999"))))

(setf *robots* (make-hash-table :test #'equalp))
(setf *packages* (make-hash-table :test #'equalp))

(defstruct robot
  pos
  items)

(defun get-robot (id)
  (gethash id *robots*))

(defun get-pkg (id)
  (gethash id *packages*))

(defun add-robot (id pos &optional dir)
  (setf (gethash id *robots*) (make-robot :pos pos)))

(defun add-pkg (id dest weight)
  (setf (gethash id *packages*) (make-pkg :dest-pos dest :weight weight)))

(defun move-robot (id dir)
  (let ((pos (robot-pos (gethash id *robots*))))
    (setf (robot-pos (gethash id *robots*)) (move pos dir))))


(defun pickup (robot p)
  (push p (robot-items (gethash robot *robots*)))
  ;[ ] todo: update package list to reflect who has it & where it was picked up
  )

(defun score (p)
  (let ((points (pkg-weight (get-pkg p))))
    (setf *score* (+ *score* points)))
  (format t "Score: ~A~%" *score*))

(defun drop (robot p)
  (format t "package: ~A" (get-pkg p))
  (format t "pos: ~A~%" *my-pos*)
  (if (equal *my-id* robot) (score p))
  (let ((i (robot-items (gethash robot *robots*))))
    (setf (robot-items (gethash robot *robots*)) (remove p i))))
    


(defun set-board (pos v &optional (board *board*))
  (setf (aref board (getx pos) (gety pos)) v))

(defun get-board (pos &optional (board *board*))
  (aref board (getx pos) (gety pos)))

(defun set-row (row str &optional (board *board*))
  (dotimes (x (- x-max 2))
    (setf (aref board (+ x 1) row) (elt str x))))

(defun blank-board (dim)
  (let ((x0 (getx dim))
	(y0 (gety dim)))
    (setf x-max (+ 2 x0))
    (setf y-max (+ 2 y0))
    (setf board (make-array (list x-max y-max) :initial-element *tile-plain*))
    (dotimes (x x-max)
      (set-board (list x 0) *tile-wall* board)
      (set-board (list x (1+ y0)) *tile-wall* board))
    (dotimes (y y-max)
      (set-board (list 0 y) *tile-wall* board)
      (set-board (list (1+ x0) y) *tile-wall* board))
    board))

(defun show-board (&optional (board *board*))
  (dotimes (y y-max)
    (dotimes (x x-max)
      (princ (get-board (list x (- y-max y 1)) board)))
    (terpri)))

(defun neighbors (pos)
  (mapcar #'(lambda (dir) (move pos dir)) '(N S E W)))

(defun lethal-neighbors (&optional (pos *my-pos*))
  (remove-if-not #'(lambda (p) (equal (get-board p) *tile-lethal*)) (neighbors pos)))

(defun read-from-server-string (str)
  (if (eql (elt str 0) #\#)
      str
    (read-from-string str)))
  

(defun parse (server-func)
  (mapcar #'read-from-server-string (tokens (funcall server-func))))

(defun make-board (server-func)
  (let ((dimension (parse server-func)))
    (let ((board (blank-board dimension))
	  (x (getx dimension))
	  (y (gety dimension)))
      (dotimes (i y)
	(set-row (+ 1 i) (funcall server-func) board))
      board)))

(defun init-environment (server-func)
  (let ((params (parse server-func)))
    (setf *my-id* (format nil "#~A" (pop params)))
    (setf *my-carrying-capacity* (pop params))
    (setf *my-money* (pop params))))

(defun show-environment ()
  (format t "Packages: ~A~%" *available-packages*)
  (format t "My-ID: ~A~%" *my-id*)
  (format t "My-Carrying-Capacity: ~A~%" *my-carrying-capacity*)
  (format t "My-Money: ~A~%" *my-money*)
  (format t "My-Route: ~A~%" *my-route*)
  (format t "My-Position: ~A~%" *my-pos*))

(defun is-dir (s)
  (position s '(N S E W)))

(defun is-robot (s)
  (and (stringp s)
       (equalp (elt s 0) #\#)))

(defun interpret-reply (l)
  (format t "Server Reply: ~A~%" l)
  (when l
    (let ((robot (pop l)))
      (let ((action (pop l)))
	(while action
	  (apply-command-to-robot robot action l)
	  (if (or (null l) (is-robot (car l)))
	      (setf action nil)
	      (setf action (pop l)))))
      (interpret-reply l))))

(defun interpret-package-list (l)
  (when l
    (let ((id (pop l)))
      (push id *available-packages*)
      (add-pkg
       id
       (list (pop l) (pop l)) ;dest
       (pop l)) ;weight
      (interpret-package-list l))))

(defun apply-command-to-robot (robot action l)
  (case action
    ((N S E W) (move-robot robot action))
    (P (pickup robot (pop l)))
    (D (drop robot (pop l)))
    (X (let ((x (pop l))
	     (junk (pop l))
	     (y (pop l)))
	 (add-robot robot (list x y))))))


(defun debug-game ()
  (defparameter *sample-board*
    '("7 5"
      "..@...."
      "......."
      "##.~~~~"
      "...~~~~"
      "......."
      "3 25 1000"
      "#2 X 2 Y 1 #1 X 1 Y 2 #3 X 1 Y 1"))
  
  
  (setf x *sample-board*)
  (setf func #'(lambda () (pop x))))

(defun plot-robots (&optional (board *board*))
  (maphash #'(lambda (id robot)
	       (set-board (robot-pos robot) #\*)) *robots*))

(defun start-game (host port)
  (setf s (socket-connect port host))
  (setf func #'(lambda ()
		 (let ((res (read-line s)))
		   (format t "Server said: ~A~%" res)
		   res)))
  (format s "Player~%")

  (setf *board* (make-board func))
  (setf *empty-board* (copy-board *board*))

  (init-environment func)
  ;;;interpret the initial server response, which assigns our position
  ;;;and that of the other robots
  (interpret-reply (parse func)))
  
(defun pre-move ()
  ;;; update global *my-pos*
  (setf *my-pos* (robot-pos (gethash *my-id* *robots*)))
   ;;;interpret the package list
  (setf *available-packages* nil)
  (interpret-package-list (parse func))

  (let ((p (position *my-pos* *base-list*)))
    (when (and p (not *available-packages*))
      (progn
	(format t "deleting unproductive base: ~A" *my-pos*)
	(delete *my-pos* *base-list* :test #'equal))))

  (plot-robots))

(defun cart-distance (pos1 pos2)
  (apply #'+ (mapcar #'abs (mapcar #'- pos1 pos2))))

(defun is-blocked (pos dir &optional (b *board*))
  (position
   (get-board (move pos dir) b)
   (list *tile-wall* *tile-lethal*)))

(defun bad-solver (pos dest)
  (setf my-board (copy-board *board*))
  (or
   (solver1 pos dest 'N my-board)
   (solver1 pos dest 'S my-board)
   (solver1 pos dest 'E my-board)
   (solver1 pos dest 'W my-board)))

(defun better-solver (pos dest)
  (let ((best nil))
    (dotimes (i 5)
      (let ((next (bad-solver pos dest)))
	(when (or (eql (length best) 0)
		  (> (length best) (length next)))
	  (setf best next))))
    best))
      
  

(defun andf (l)
  (cond ((atom l) l)
	(t (and (car l) (andf (cdr l))))))

(defun scramble (l)
  (let ((temp nil)
	(len (length l)))
    (dotimes (i (* 2 len))
      (let ((x (random len))
	    (y (random len)))
	(when (not (eql x y))
	  (progn
	    (setf temp (elt l x))
	    (setf (elt l x) (elt l y))
	    (setf (elt l y) temp)))))
    l))

(defun solver1 (pos dest dir my-board)
  (cond ((equal pos dest) (list dir))
	((is-blocked pos dir) nil)
	((equal #\V (get-board (move pos dir) my-board)) nil)
	(t (set-board (move pos dir) #\V my-board)
	   (let ((solution nil))
	     (dolist (d (scramble '(N S E W)))
	       (let ((next-try (solver1 (move pos dir) dest d my-board)))
		 (when (and next-try) (setf solution (cons dir next-try)))))
	     solution))))
	       

(defun perform-move (bid move)
  (setf *my-money* (- *my-money* (abs bid)))
  (format s "~A ~A~%" bid move)
  (format nil "My Move: ~A ~A~%" bid move)
  (setf *board* (copy-board *empty-board*))
  (interpret-reply (parse func)))
;;(show-environment))


;(dotimes (i 10)
;  (pre-move)
;  (perform-move 1 "Move E"))
;(dotimes (i 10)
;  (pre-move)
;  (perform-move 1 "Move N"))

(defun my-packages ()
  (robot-items (get-robot *my-id*)))

(defun get-dest-pos (id)
  (let ((p (get-pkg id)))
    (when p (pkg-dest-pos p))))

(defun packages-i-can-deliver-here (&optional (pos *my-pos*))
  (remove-if-not #'(lambda (x) (equal (get-dest-pos x) pos))
		 (my-packages)))

(defun my-remaining-carrying-capacity ()
  (- *my-carrying-capacity* (apply #'+ (mapcar #'(lambda (p) (pkg-weight (get-pkg p))) (my-packages)))))

(defun packages-here-that-i-can-carry (&optional (pos *my-pos*))
  (let ((w (my-remaining-carrying-capacity)))
    (remove-if #'(lambda (p) (> (pkg-weight (get-pkg p)) w))
		   *available-packages*)))
  

(defun list-to-string (l)
  (string-trim " "
	       (with-output-to-string (s)
				      (dolist (i l)
					(format s "~A " i s)))))

(defun drop-off-packages (l)
  (format nil "Drop ~A" (list-to-string l)))

(defun pickup-packages (l)
  (format nil "Pick ~A" (list-to-string l)))

(defun move-myself (d)
  (format nil "Move ~A" d))

(defun packages-destined-for (pos)
  (remove-if-not #'(lambda (p) (equalp (pkg-dest-pos (get-pkg p)) pos))
	     (my-packages)))

(defun pick-one (l)
  (elt l (random (length l))))

(defun go-there ()
  (setf *my-route* (make-route *my-pos* (pick-one *base-list*))))

(defun i-know-where-package-might-be ()
  *base-list*)

(defun random-location ()
  (let* ((pos (list (random x-max) (random y-max)))
	 (contents (get-board pos)))
    (if (position contents '(*tile-lethal* *tile-wall*))
	(random-location) ;try again
      pos)))
   

(defun choose-random-destination ()
  (setf *my-route* (make-route *my-pos* (random-location)))
  (continue-route))

(defun pick-good-package-destination (l) ;;; good? how about 1st?
  (pkg-dest-pos (get-pkg (car l))))
  

(defun strategy1 ()
  (let ((move nil)
	(bid 1))
    (pre-move)
    
    ;;; first, either drop-off or pickup-packages
    (aif (packages-i-can-deliver-here)
	 (setf move (drop-off-packages it))
	 (aif (packages-here-that-i-can-carry)
	      (setf move (pickup-packages it))))

    ;;; or continue delivering packages:
    (when (and (not move) (my-packages)) ; I have packages
      (setf move
	    (if (packages-destined-for (route-dest *my-route*))
		(continue-route)
	      (progn
		(setf *my-route* (make-route *my-pos* (pick-good-package-destination (my-packages))))
		(continue-route)))))

    ;;; or continue doing whatever
    (when (not move)
      (setf move 
	    (cond (*my-route* (continue-route))
		  ((i-know-where-package-might-be) (go-there))
		  (t (choose-random-destination)))))

    
    ;(when (my-proposed-move-puts-me-in-danger)
    ;  (setf move (stay-still)))
  
    ;(when (and
;   ;   (i-plan-to-move)
	;   (moving-last-puts-me-in-danger-of-being-pushed-into-water)
	 ;  (i-have-plenty-of-money))
      ;(raise-bid))
    (format t "preparing to move: ~A~%" move)

    (when (lethal-neighbors)
      (incf bid (random 5)))
    
    (perform-move bid move)))


;;; defun plot-robots - for each robot, plot it on the map before
;;; displaying it. Show ME as a different symbol than the rest

;;; wall-hugger - If No wall in "front" of me, go straight; else "turn"
;;; left and repeat (start with front=North, then a left turn causes it
;;; to be East, then South, then West, then North --- whenever a package
;;; is encountered, pick it up



(defun play-now (host port)
  (start-game host port)
  (loop
   (strategy1)))

